/* autonosim_mpi.c : MPI message passing */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autonosim.h"

#include <mpi.h>



#if 0
#define DEBUG_MSG
#endif



/* Initialization and Termination */

static int NTasks;
static int MyTaskId;

void AutoNOSim_Initialize (int *argc, char ***argv)
{
  MPI_Init (argc, argv);
  MPI_Comm_rank (MPI_COMM_WORLD, &MyTaskId);
  MPI_Comm_size (MPI_COMM_WORLD, &NTasks);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] Initialize OK NTasks %d \n", 
	   MyTaskId, NTasks);
#endif
}

void AutoNOSim_Terminate (void)
{
  MPI_Finalize ();

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] Terminate OK \n", 
	   MyTaskId);
#endif
}

int AutoNOSim_NTasks (void)
{
  return NTasks;
}

int AutoNOSim_MyTaskId (void)
{
  return MyTaskId;
}



/* General Communication */

#define BUFFER_SIZE 40000000

static int IsPacking;
static int PackingPosition;
static int UnpackingPosition;
static char Buffer[BUFFER_SIZE];

void AutoNOSim_Clear (int isPacking)
{
  IsPacking = isPacking;
  PackingPosition = 0;
  UnpackingPosition = 0;
}

int AutoNOSim_IsPacking (void)
{
  return IsPacking;
}

void AutoNOSim_Integer (int *value)
{
  if (IsPacking) {
    assert(PackingPosition + sizeof (int) < BUFFER_SIZE);
    MPI_Pack (value, 1, MPI_INT, 
	      Buffer, BUFFER_SIZE, &PackingPosition, 
	      MPI_COMM_WORLD);
  } else {
    assert(UnpackingPosition + sizeof (int) < BUFFER_SIZE);
    MPI_Unpack (Buffer, BUFFER_SIZE, &UnpackingPosition, 
		value, 1, MPI_INT, 
		MPI_COMM_WORLD);
  }
} 

void AutoNOSim_Integers (int values[], int size)
{
  if (IsPacking) {
    assert(PackingPosition + sizeof (int) * size < BUFFER_SIZE);
    MPI_Pack (values, size, MPI_INT, 
	      Buffer, BUFFER_SIZE, &PackingPosition, 
	      MPI_COMM_WORLD);
  } else {
    assert(UnpackingPosition + sizeof (int) * size < BUFFER_SIZE);
    MPI_Unpack (Buffer, BUFFER_SIZE, &UnpackingPosition, 
		values, size, MPI_INT, 
		MPI_COMM_WORLD);
  }
} 

void AutoNOSim_Real (double *value)
{
  if (IsPacking) {
    assert(PackingPosition + sizeof (double) < BUFFER_SIZE);
    MPI_Pack (value, 1, MPI_DOUBLE, 
	      Buffer, BUFFER_SIZE, &PackingPosition, 
	      MPI_COMM_WORLD);
  } else {
    assert(UnpackingPosition + sizeof (double) < BUFFER_SIZE);
    MPI_Unpack (Buffer, BUFFER_SIZE, &UnpackingPosition, 
		value, 1, MPI_DOUBLE, 
		MPI_COMM_WORLD);
  }
}

void AutoNOSim_Reals (double values[], int size)
{
  if (IsPacking) {
    assert(PackingPosition + sizeof (double) * size < BUFFER_SIZE);
    MPI_Pack (values, size, MPI_DOUBLE, 
	      Buffer, BUFFER_SIZE, &PackingPosition, 
	      MPI_COMM_WORLD);
  } else {
    assert(UnpackingPosition + sizeof (double) * size < BUFFER_SIZE);
    MPI_Unpack (Buffer, BUFFER_SIZE, &UnpackingPosition, 
		values, size, MPI_DOUBLE, 
		MPI_COMM_WORLD);
  }
}

void AutoNOSim_String (char values[], int size)
{
  if (IsPacking) {
    assert(PackingPosition + sizeof (char) * size < BUFFER_SIZE);
    MPI_Pack (values, size, MPI_CHAR, 
	      Buffer, BUFFER_SIZE, &PackingPosition, 
	      MPI_COMM_WORLD);
  } else {
    assert(UnpackingPosition + sizeof (char) * size < BUFFER_SIZE);
    MPI_Unpack (Buffer, BUFFER_SIZE, &UnpackingPosition, 
		values, size, MPI_CHAR, 
		MPI_COMM_WORLD);
  }
}

void AutoNOSim_SendTo (int taskId)
{
  if (taskId != MyTaskId) {
    MPI_Send (Buffer, PackingPosition, MPI_PACKED, 
	      taskId, 0, MPI_COMM_WORLD);
  }
}

void AutoNOSim_ReceiveFrom (int taskId)
{
  if (taskId != MyTaskId) {
    MPI_Status status;

    MPI_Recv (Buffer, sizeof (Buffer), MPI_PACKED, 
	      taskId, 0, MPI_COMM_WORLD, &status);
  }
}  

int AutoNOSim_ReceiveFromAny (void)
{
  MPI_Status status;
  int taskId;

  MPI_Recv (Buffer, sizeof (Buffer), MPI_PACKED, 
	    MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &status);
  taskId = status.MPI_SOURCE;
  return taskId;
}  

void AutoNOSim_SendReceive (int taskId)
{
  static char buffer2[BUFFER_SIZE];
  int i;

  if (taskId != MyTaskId) {
    MPI_Status status;
    
    for (i = 0; i < PackingPosition; i++) {
      buffer2[i] = Buffer[i];
    }

    MPI_Sendrecv (buffer2, PackingPosition, MPI_PACKED, 
		  taskId, 0,
		  Buffer, sizeof (Buffer), MPI_PACKED, 
		  taskId, 0, 
		  MPI_COMM_WORLD, &status);
    
  }
}  

void AutoNOSim_Barrier (void)
{
  MPI_Barrier (MPI_COMM_WORLD);
}  

void AutoNOSim_GetSum 
(double *sum_IO)
{
  double result;
  double domainSum;

  domainSum = *sum_IO;

  MPI_Allreduce (&domainSum, &result, 1, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

  *sum_IO = result;
}

void AutoNOSim_GetSum2
(double *sum0_IO, double *sum1_IO)
{
  double results[2];
  double domainSums[2];

  domainSums[0] = *sum0_IO;
  domainSums[1] = *sum1_IO;

  MPI_Allreduce (domainSums, results, 2, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

  *sum0_IO = results[0];
  *sum1_IO = results[1];
}

void AutoNOSim_GetMin 
(double *min_IO)
{
  double result;
  double domainMin;

  domainMin = *min_IO;

  MPI_Allreduce (&domainMin, &result, 1, 
		 MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);

  *min_IO = result;
}

void AutoNOSim_GetMax 
(double *max_IO)
{
  double result;
  double domainMax;

  domainMax = *max_IO;

  MPI_Allreduce (&domainMax, &result, 1, 
		 MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

  *max_IO = result;
}



/* Data Exchange Across Domain Boundary */

#define MAX_COMPONENTS 6
#define MAX_ENTITYS 1000000
#define MAX_BOUNDARYS 50
#define MAX_BOUNDARY_ENTITYS 10000

static int NEntitys;
static int NGlobalEntitys;

static int Entity_IsExcluded[MAX_ENTITYS];

static int Entity_IsOnBoundary[MAX_ENTITYS];

static int NBoundarys;

static int Boundary_NeighborDomainIds[MAX_BOUNDARYS];

static int Boundary_NEntitys[MAX_BOUNDARYS];

static int Boundary_EntityIds[MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS];
/* Domain-local entity ID */
/* They are sorted by their corresponding global entity ID. */

static double Boundary_ValuesToBeSent
[MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS * MAX_COMPONENTS];
static double Boundary_ValuesReceived
[MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS * MAX_COMPONENTS];
/* Send-receive buffer of exchanged entity values */

static MPI_Request Requests[MAX_BOUNDARYS * 2];

static int RequestIsDone[MAX_BOUNDARYS * 2];

void AutoNOSim_ReadExchangeFile (char *baseFileName)
{
  int domainId;
  int nDomains;
  char fileName[256];
  FILE* fp;
  int nExcludedEntitys;
  int i;
  int iBoundary;
  int iEntity;
  
  domainId = MyTaskId;
  nDomains = NTasks;
  
  sprintf (fileName, "%s.%d", baseFileName, domainId);
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", &NEntitys);
  assert(NEntitys <= MAX_ENTITYS);

  fscanf (fp, "%d", &NGlobalEntitys);

  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    Entity_IsExcluded[iEntity] = 0;
  }

  fscanf (fp, "%d", &nExcludedEntitys);
  
  for (i = 0; i < nExcludedEntitys; i++) {
    int entityId;

    fscanf (fp, "%d", &entityId);
    assert(0 <= entityId);
    assert(entityId < NEntitys);
    Entity_IsExcluded[entityId] = 1;
  }

  fscanf (fp, "%d", &NBoundarys);
  assert(NBoundarys <= MAX_BOUNDARYS);
  
  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    fscanf (fp, "%d", &Boundary_NeighborDomainIds[iBoundary]);
    assert(0 <= Boundary_NeighborDomainIds[iBoundary]);
    assert(Boundary_NeighborDomainIds[iBoundary] < nDomains);

    fscanf (fp, "%d", &Boundary_NEntitys[iBoundary]);
    assert(Boundary_NEntitys[iBoundary] <= MAX_BOUNDARY_ENTITYS);
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      fscanf (fp, "%d", &Boundary_EntityIds[iBoundary][iEntity]);
      assert(0 <= Boundary_EntityIds[iBoundary][iEntity]);
      assert(Boundary_EntityIds[iBoundary][iEntity] < NEntitys);
    }
  }
  
  fclose (fp);

  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    Entity_IsOnBoundary[iEntity] = 0;
  }

  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      int entityId = Boundary_EntityIds[iBoundary][iEntity];

      Entity_IsOnBoundary[entityId] = 1;
    }
  }
}

int AutoNOSim_NEntitys ()
{
  return NEntitys;
}

int AutoNOSim_NGlobalEntitys ()
{
  return NGlobalEntitys;
}

int AutoNOSim_EntityIsExcluded (int iEntity)
{
  assert(0 <= iEntity);
  assert(iEntity < NEntitys);
  
  return Entity_IsExcluded[iEntity];
}

int AutoNOSim_EntityIsOnBoundary (int iEntity)
{
  assert(0 <= iEntity);
  assert(iEntity < NEntitys);
  
  return Entity_IsOnBoundary[iEntity];
}

static void PrepareSendingValues (int nComponents, double values[])
{
  int iBoundary;
  int iEntity;
  int iComponent;

  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {

    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      int entityId = Boundary_EntityIds[iBoundary][iEntity];
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	Boundary_ValuesToBeSent
	  [iBoundary][iEntity * nComponents + iComponent] 
	  = values[entityId * nComponents + iComponent];
      }
    }
  }
}

static void SendReceiveBoundaryValues (int nComponents)
{
  int iBoundary;
  int iRequest;
  int count;
  
  count = 0;
  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    int destTaskId = Boundary_NeighborDomainIds[iBoundary];
    int nEntitys = Boundary_NEntitys[iBoundary];
    
    MPI_Isend (Boundary_ValuesToBeSent[iBoundary], 
	       nEntitys * nComponents, MPI_DOUBLE, 
	       destTaskId, 0, MPI_COMM_WORLD, 
	       &Requests[count]);
    count++;
    MPI_Irecv (Boundary_ValuesReceived[iBoundary], 
	       nEntitys * nComponents, MPI_DOUBLE, 
	       destTaskId, 0, MPI_COMM_WORLD, 
	       &Requests[count]);
    count++;
  }

  for (iRequest = 0; iRequest < NBoundarys * 2; iRequest++) {
    RequestIsDone[iRequest] = 0;
  }
}

static void CheckRequests ()
{
  int iRequest;
  int flag;
  MPI_Status status;

  for (iRequest = 0; iRequest < NBoundarys * 2; iRequest++) {
    if (RequestIsDone[iRequest]) continue;

    flag = 0;
    MPI_Test (&Requests[iRequest], &flag, &status);
    if (flag) {
      RequestIsDone[iRequest] = 1;
    }
  }
}

static void WaitRequests ()
{
  static MPI_Status statuss[MAX_BOUNDARYS * 2];

  MPI_Waitall (NBoundarys * 2, Requests, statuss);
}

static void PerformOperationOnValues 
(int operation, int nComponents, double values[])
{
  int iBoundary;
  int iEntity;
  int iComponent;

  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      int entityId = Boundary_EntityIds[iBoundary][iEntity];
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	double valueReceived = Boundary_ValuesReceived
	  [iBoundary][iEntity * nComponents + iComponent];
	
	switch (operation) {
	case AUTONOSIM_OP_SUM:
	  values[entityId * nComponents + iComponent] += valueReceived;
	  break;
	case AUTONOSIM_OP_PRODUCT:
	  values[entityId * nComponents + iComponent] *= valueReceived;
	  break;
	case AUTONOSIM_OP_MIN:
	  {
	    double value = values[entityId * nComponents + iComponent];
	    
	    values[entityId * nComponents + iComponent] 
	      = (value < valueReceived) ? value : valueReceived;
	  }
	  break;
	case AUTONOSIM_OP_MAX:
	  {
	    double value = values[entityId * nComponents + iComponent];
	    
	    values[entityId * nComponents + iComponent] 
	      = (value < valueReceived) ? valueReceived : value;
	  }
	  break;
	default:
	  assert(0);
	  break;
	}
      }
    }
  }
}  

void AutoNOSim_Exchange 
(int operation, int nComponents, double values[])
{
  assert(nComponents <= MAX_COMPONENTS);

  PrepareSendingValues (nComponents, values);
  SendReceiveBoundaryValues (nComponents);
  WaitRequests ();
  PerformOperationOnValues (operation, nComponents, values);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] exchange ( operationID %d ) \n", 
	   MyTaskId, operation);
#endif
}

void AutoNOSim_PreExchange 
(int nComponents, double values[])
{
  assert(nComponents <= MAX_COMPONENTS);

  PrepareSendingValues (nComponents, values);
  SendReceiveBoundaryValues (nComponents);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] pre-exchange \n", 
	   MyTaskId);
#endif
}

void AutoNOSim_CheckExchange (void)
{
  CheckRequests ();
}

void AutoNOSim_PostExchange 
(int operation, int nComponents, double values[])
{
  assert(nComponents <= MAX_COMPONENTS);

  WaitRequests ();
  PerformOperationOnValues (operation, nComponents, values);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] post-exchange ( operationID %d ) \n", 
	   MyTaskId, operation);
#endif
}

double AutoNOSim_ScalarProduct 
(int nComponents, double values0[], double values1[])
{
  double result;
  double domainSum;
  int iEntity;
  int iComponent;

  domainSum = 0.0;
  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    if (!Entity_IsExcluded[iEntity]) {
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	domainSum += values0[iEntity * nComponents + iComponent]
	  * values1[iEntity * nComponents + iComponent];
      }
    }
  }

  MPI_Allreduce (&domainSum, &result, 1, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] scalar product \n", 
	   MyTaskId);
#endif

  return result;
}

void AutoNOSim_GetTwoScalarProducts 
(int nComponents, 
 double values0[], double values1[], double *scalarProductOfValues01, 
 double values2[], double values3[], double *scalarProductOfValues23)
{
  double results[2];
  double domainSums[2];
  int iEntity;
  int iComponent;

  domainSums[0] = 0.0;
  domainSums[1] = 0.0;
  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    if (!Entity_IsExcluded[iEntity]) {
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	domainSums[0] += values0[iEntity * nComponents + iComponent]
	  * values1[iEntity * nComponents + iComponent];
	domainSums[1] += values2[iEntity * nComponents + iComponent]
	  * values3[iEntity * nComponents + iComponent];
      }
    }
  }

  MPI_Allreduce (domainSums, results, 2, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] scalar product \n", 
	   MyTaskId);
#endif

  *scalarProductOfValues01 = results[0];
  *scalarProductOfValues23 = results[1];
}
